@@ -9,14 +9,19 @@
  * the HTTP headers to avoid DNS lookups.
  */
 
+#include <sys/types.h>
 #include <sys/time.h>
 #include <sys/socket.h>
+#include <sys/stat.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
 #include <unistd.h>
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 #include <syslog.h>
+#endif
 #include <signal.h>
+#include <fcntl.h>
 #include <pwd.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -44,6 +49,16 @@
 #define SIGCHLD SIGCLD
 #endif
 
+#define BUFFER_SIZE			(32769)
+
+#define IN_METHOD			0
+#define IN_URL				1
+#define IN_SKIP_URL			2
+#define IN_VERSION			3
+#define IN_ALMOST_HEADER	4
+#define IN_HEADER			5
+#define IN_SKIP_CONNECTION	6
+
 /*
  * Typedefs.
  */
@@ -53,6 +68,7 @@ typedef unsigned long	ipaddr_t;
  * Macros.
  */
 #define FD_MAX(a,b)		((a) > (b) ? (a) : (b))
+#define get_end_pos(b)	strlen(b)
 
 /*
  * Function prototypes.
@@ -66,18 +82,21 @@ static int bind_to_port(ipaddr_t bind_ip, short bind_port);
 static int connect_to_proxy(ipaddr_t ip, short port);
 static void lookup_hostname(struct sockaddr_in *addr, char *hostname,
 							int hostlen, int needed);
-static void server_main_loop(int sock, ipaddr_t server_ip, short server_port);
+static void server_main_loop(int sock, char *server_hostname, short server_port);
 static void trans_proxy(int sock, struct sockaddr_in *addr,
-						ipaddr_t server_ip, short server_port);
+						char *server_hostname, short server_port);
 static int	get_site_from_headers(int sock, char *hostname, int hostlen,
 								  char *headers, int header_max,
 								  int *header_len);
 static void alarm_signal(int sig);
+static void append_string(char *dest, int max_size, char *source);
+static void append_char(char *dest, int max_size, char source);
 
 /*
  * Command line switches.
  */
 static int				fully_transparent = 0;
+static char				*force_url = NULL;
 
 #ifdef IPFILTER
 /*
@@ -95,7 +114,7 @@ int main(int argc, char **argv)
 	int					run_as_server = 0;
 	ipaddr_t			bind_ip = INADDR_ANY;
 	short				bind_port = -1;
-	ipaddr_t			server_ip = INADDR_NONE;
+	char 				*server_hostname = NULL;
 	short				server_port = -1;
 	uid_t				run_uid = -1;
 	gid_t				run_gid = -1;
@@ -106,7 +125,7 @@ int main(int argc, char **argv)
 	/*
 	 * Parse the command line arguments.
 	 */
-	while ((arg = getopt(argc, argv, "ts:r:b:")) != EOF)
+	while ((arg = getopt(argc, argv, "ts:r:b:f:")) != EOF)
 	{
 		switch (arg)
 		{
@@ -128,6 +147,10 @@ int main(int argc, char **argv)
 			bind_ip = getipaddress(optarg);
 			break;
 
+		case 'f':
+			force_url = optarg;
+			break;
+
 		case '?':
 			usage(argv[0], NULL);
 			break;
@@ -139,9 +162,9 @@ int main(int argc, char **argv)
 	 */
 	for (; optind < argc; ++optind)
 	{
-		if (server_ip == INADDR_NONE)
+		if (server_hostname == NULL)
 		{
-			server_ip = getipaddress(argv[optind]);
+			server_hostname = argv[optind];
 		}
 		else if (server_port == -1)
 		{
@@ -156,7 +179,7 @@ int main(int argc, char **argv)
 	/*
 	 * Test to make sure required args were provided and are valid.
 	 */
-	if (server_ip == INADDR_NONE)
+	if (server_hostname == NULL)
 	{
 		usage(argv[0], "No proxyhost was specified (or it was invalid).");
 	}
@@ -209,10 +232,12 @@ int main(int argc, char **argv)
 			exit(1);
 
 		case 0:
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 			/*
 			 * Open syslog for logging errors.
 			 */
-			openlog("tproxy", LOG_PID, LOG_DAEMON);
+			openlog(argv[0], LOG_PID, LOG_DAEMON);
+#endif
 
 			/*
 			 * Ignore some signals.
@@ -241,12 +266,14 @@ int main(int argc, char **argv)
 			/*
 			 * Handle the server main loop.
 			 */
-			server_main_loop(sock, server_ip, server_port);
+			server_main_loop(sock, server_hostname, server_port);
 
 			/*
 			 * Should never exit.
 			 */
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 			closelog();
+#endif
 			exit(1);
 		}
 
@@ -256,10 +283,12 @@ int main(int argc, char **argv)
 		exit(0);
 	}
 
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 	/*
 	 * Open syslog for logging errors.
 	 */
-	openlog("tproxy", LOG_PID, LOG_DAEMON);
+	openlog(argv[0], LOG_PID, LOG_DAEMON);
+#endif
 
 	/*
 	 * We are running from inetd so find the peer address.
@@ -267,9 +296,15 @@ int main(int argc, char **argv)
 	len = sizeof(addr);
 	if (getpeername(STDIN_FILENO, (struct sockaddr *)&addr, &len) < 0)
 	{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 		syslog(LOG_ERR, "getpeername(): %m");
+#ifndef DEBUG
 		closelog();
+#endif
+#endif
+#ifndef DEBUG
 		exit(1);
+#endif
 	}
 
 	/*
@@ -283,8 +318,10 @@ int main(int argc, char **argv)
 	/*
 	 * We are running from inetd so process stdin.
 	 */
-	trans_proxy(STDIN_FILENO, &addr, server_ip, server_port);
+	trans_proxy(STDIN_FILENO, &addr, server_hostname, server_port);
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 	closelog();
+#endif
 
 	return (0);
 }
@@ -298,8 +335,9 @@ static void usage(char *prog, char *opt)
 	{
 		fprintf(stderr, "%s: %s\n", prog, opt);
 	}
-	fprintf(stderr, "usage: %s [-t] [-s port [-r user] [-b ipaddr]] proxyhost proxyport\n", prog);
+	fprintf(stderr, "usage: %s [-t] [-f url] [-s port [-r user] [-b ipaddr]] proxyhost proxyport\n", prog);
 	fprintf(stderr, "    -t          Act fully transparently (default is HTTP translate).\n");
+	fprintf(stderr, "    -f url      Force access to always go to specified URL.\n");
 	fprintf(stderr, "    -s port     Run as a server bound to the specified port.\n");
 	fprintf(stderr, "    -r user     Run as the specified user in server mode.\n");
 	fprintf(stderr, "    -b ipaddr   Bind to the specified ipaddr in server mode.\n");
@@ -492,7 +530,9 @@ static int connect_to_proxy(ipaddr_t ip, short port)
 	 */
 	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
 	{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 		syslog(LOG_WARNING, "socket(): %m");
+#endif
 		return (-1);
 	}
 
@@ -509,7 +549,9 @@ static int connect_to_proxy(ipaddr_t ip, short port)
 	 */
 	if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)
 	{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 		syslog(LOG_WARNING, "connect(): %m");
+#endif
 		return (-1);
 	}
 
@@ -551,7 +593,9 @@ static void lookup_hostname(struct sockaddr_in *addr, char *hostname,
 		}
 		else
 		{
+#ifdef LOG_TO_SYSLOG
 			syslog(LOG_INFO, "DNS lookup for %s failed: %m", hostname);
+#endif
 		}
 	}
 #ifdef USELESS_DNS_LOOKUPS
@@ -565,7 +609,9 @@ static void lookup_hostname(struct sockaddr_in *addr, char *hostname,
 		}
 		else
 		{
+#ifdef LOG_TO_SYSLOG
 			syslog(LOG_INFO, "DNS lookup for %s failed: %m", hostname);
+#endif
 		}
 	}
 #endif
@@ -575,7 +621,7 @@ static void lookup_hostname(struct sockaddr_in *addr, char *hostname,
 /*
  * This is the main loop when running as a server.
  */
-static void server_main_loop(int sock, ipaddr_t server_ip, short server_port)
+static void server_main_loop(int sock, char *server_hostname, short server_port)
 {
 	int					new_sock;
 	struct sockaddr_in	addr;
@@ -608,7 +654,14 @@ static void server_main_loop(int sock, ipaddr_t server_ip, short server_port)
 			/*
 			 * Connection resets are common enough to log them as debug only.
 			 */
-			syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_ERR), "accept(): %m");
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+			if (errno != ECONNRESET)
+				syslog(LOG_ERR, "accept(): %m");
+# ifdef LOG_TO_SYSLOG
+			else
+				syslog(LOG_DEBUG, "accept(): %m");
+# endif
+#endif
 		}
 
 		/*
@@ -626,9 +679,11 @@ static void server_main_loop(int sock, ipaddr_t server_ip, short server_port)
 			/*
 			 * Start the proxy work in the new socket.
 			 */
-			trans_proxy(new_sock, &addr, server_ip, server_port);
+			trans_proxy(new_sock, &addr, server_hostname, server_port);
 			close(new_sock);
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 			closelog();
+#endif
 			exit(0);
 		}
 
@@ -643,17 +698,22 @@ static void server_main_loop(int sock, ipaddr_t server_ip, short server_port)
  * Perform the transparent proxy activity.
  */
 static void trans_proxy(int sock, struct sockaddr_in *from_addr,
-						ipaddr_t server_ip, short server_port)
+						char *server_hostname, short server_port)
 {
+	ipaddr_t			server_ip;
 	struct sockaddr_in	to_addr;
 	int					to_len;
 	int					proxy;
 	char				from_host[64];
 	char				to_host[64];
-	static char			headers[16384 + 1];
+	static char			headers[BUFFER_SIZE];
 	int					header_len;
-	char				*uri_pos;
-	char				*uri_end;
+	static char			proxy_request[BUFFER_SIZE];
+	int					mode;
+	char				*current_pos;
+	int					url_start = 0;
+	int					url_end = 0;
+	int					translate_url = 0;
 	int					max_fd;
 	fd_set				read_fd;
 	int					read_len;
@@ -661,6 +721,19 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 	natlookup_t			natlook;
 #endif
 
+	/*
+	 * Lookup proxy server IP address. Because we do the lookup for every
+	 * request we make use of DNS round-robin for load balancing.
+	 */
+	server_ip = getipaddress(server_hostname);
+	if (server_ip == INADDR_NONE)
+	{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+		syslog(LOG_ERR, "Could not resolve IP address for proxy: %s", server_hostname);
+#endif
+		return;
+	}
+
 	/*
 	 * The first thing we do is get the IP address that the client was
 	 * trying to connected to. Here lies part of the magic. Normally
@@ -669,8 +742,12 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 	to_len = sizeof(to_addr);
 	if (getsockname(sock, (struct sockaddr *)&to_addr, &to_len) < 0)
 	{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 		syslog(LOG_ERR, "getsockname(): %m");
+#endif
+#ifndef DEBUG
 		return;
+#endif
 	}
 
 #ifdef IPFILTER
@@ -696,7 +773,9 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 	 */
 	if (ioctl(natdev, SIOCGNATL, &natlook) == -1)
 	{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 		syslog(LOG_ERR, "ioctl(SIOCGNATL): %m");
+#endif
 		return;
 	}
 
@@ -725,7 +804,9 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 		/*
 		 * Log the facts about the connection.
 		 */
+#ifdef LOG_TO_SYSLOG
 		syslog(LOG_INFO, "Transparent %s -> %s", from_host, to_host);
+#endif
 	}
 	else
 	{
@@ -737,17 +818,20 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 		/*
 		 * See if the destination site can be found in the HTTP headers.
 		 */
-		if (get_site_from_headers(sock, to_host, sizeof(to_host),
+		switch (get_site_from_headers(sock, to_host, sizeof(to_host),
 								  headers, sizeof(headers) - 1,
 								  &header_len))
 		{
+		case 1:
 			/*
 			 * Log the facts about the connection.
 			 */
+#ifdef LOG_TO_SYSLOG
 			syslog(LOG_INFO, "Request NoDNS %s -> %s", from_host, to_host);
-		}
-		else
-		{
+#endif
+			break;
+
+		case 0:
 			/*
 			 * Fallback is to use the DNS to find the destination site.
 			 */
@@ -756,62 +840,174 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 			/*
 			 * Log the facts about the connection.
 			 */
+#ifdef LOG_TO_SYSLOG
 			syslog(LOG_INFO, "Request %s -> %s", from_host, to_host);
+#endif
+			break;
+
+		case -1:
+			close(proxy);
+			return;
 		}
 
-#ifdef PAY_THE_PENALTY
+#ifdef DEBUG_INPUT_REQUESTS
 		/*
-		 * Not that I'm mean or anything, but if they don't configure to
-		 * use the proxy directly then they pay a little penalty.
+		 * Log the request headers.
 		 */
-		sleep(2);
+#ifdef LOG_TO_SYSLOG
+		syslog(LOG_NOTICE, "Request In: '%s'", headers);
 #endif
 
-		/*
-		 * Parse the request line which is of the form METHOD /URI HTTP/1.0
-		 * We split the line between the METHOD and /URI. If the URI starts
-		 * with a / then we have transparently trapped a request, if it doesn't
-		 * then assume that we have been nominated as a proxy by the browser.
-		 */
-		if (((uri_pos = strchr(headers, ' ')) != NULL) &&
-			(uri_pos[1] == '/'))
-		{
-			*uri_pos++ = '\0';
-			write(proxy, headers, strlen(headers));
-			write(proxy, " http://", 8);
-			write(proxy, to_host, strlen(to_host));
-			write(proxy, uri_pos, header_len - strlen(headers) - 1);
-
-			/*
-			 * Dump a nice log message showing what we have done.
-			 */
-			if ((uri_end = strchr(uri_pos, ' ')) != NULL)
-			{
-				*uri_end = '\0';
-				syslog(LOG_DEBUG, "Translated '%s http://%s%s'", headers, to_host, uri_pos);
 #ifdef DEBUG
-				fprintf(stderr, "Translated '%s http://%s%s'\n", headers, to_host, uri_pos);
+		fprintf(stderr, "Request In: '%s'\n", headers);
 #endif
+#endif
+
+		/*
+		 * Parse the request which is of the form...
+		 *     METHOD [proto://host]/URI[ HTTP/1.0]\r\n
+		 *     Header: data\r\n
+		 *     ...
+		 *     \r\n
+		 *
+		 * We use a state machine to parse the request.
+		 */
+		mode = IN_METHOD;
+		for (current_pos = headers; *current_pos; ++current_pos)
+		{
+			switch (mode)
+			{
+			case IN_METHOD:
+				append_char(proxy_request, BUFFER_SIZE, *current_pos);
+				url_start = get_end_pos(proxy_request);
+				if (*current_pos == ' ')
+				{
+					if (force_url)
+					{
+						append_string(proxy_request, BUFFER_SIZE, force_url);
+						url_end = get_end_pos(proxy_request);
+						mode = IN_SKIP_URL;
+					}
+					else
+					{
+						if (current_pos[1] == '/')
+						{
+							translate_url = 1;
+							append_string(proxy_request, BUFFER_SIZE, "http://");
+							append_string(proxy_request, BUFFER_SIZE, to_host);
+						}
+						mode = IN_URL;
+					}
+				}
+				break;
+
+			case IN_URL:
+				if (*current_pos == ' ')
+				{
+					url_end = get_end_pos(proxy_request);
+					mode = IN_VERSION;
+				}
+				else if (*current_pos == '\r')
+				{
+					url_end = get_end_pos(proxy_request);
+					mode = IN_ALMOST_HEADER;
+				}
+				append_char(proxy_request, BUFFER_SIZE, *current_pos);
+				break;
+
+			case IN_SKIP_URL:
+				if (*current_pos == ' ')
+				{
+					append_char(proxy_request, BUFFER_SIZE, *current_pos);
+					mode = IN_VERSION;
+				}
+				else if (*current_pos == '\r')
+				{
+					append_char(proxy_request, BUFFER_SIZE, *current_pos);
+					mode = IN_ALMOST_HEADER;
+				}
+				break;
+
+			case IN_VERSION:
+				append_char(proxy_request, BUFFER_SIZE, *current_pos);
+				if (*current_pos == '\r')
+					mode = IN_ALMOST_HEADER;
+				break;
+
+			case IN_ALMOST_HEADER:
+				append_char(proxy_request, BUFFER_SIZE, *current_pos);
+				if (*current_pos == '\n')
+					mode = IN_HEADER;
+				break;
+
+			case IN_HEADER:
+				if (*current_pos == '\r')
+				{
+					append_char(proxy_request, BUFFER_SIZE, *current_pos);
+					mode = IN_ALMOST_HEADER;
+				}
+				else if (strncasecmp(current_pos, "Connection:", 11) == 0)
+					mode = IN_SKIP_CONNECTION;
+				else
+					append_char(proxy_request, BUFFER_SIZE, *current_pos);
+				break;
+			
+			case IN_SKIP_CONNECTION:
+				if (*current_pos == '\n')
+					mode = IN_HEADER;
+				break;
 			}
 		}
+
+		if (write(proxy, proxy_request, get_end_pos(proxy_request)) < 0)
+		{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+			syslog(LOG_WARNING, "write(proxy) failed: %m");
+#endif
+			close(proxy);
+			return;
+		}
+
+#ifdef DEBUG_OUTPUT_REQUESTS
+#ifdef LOG_TO_SYSLOG
+		syslog(LOG_NOTICE, "Request Out: '%s'", proxy_request);
+#endif
+
+#ifdef DEBUG
+		fprintf(stderr, "Request Out: '%s'\n", proxy_request);
+#endif
+#endif
+
+		/*
+		 * Dump a nice log message showing what we have done.
+		 */
+		proxy_request[url_end] = '\0';
+		if (force_url)
+		{
+#ifdef LOG_TO_SYSLOG
+			syslog(LOG_NOTICE, "Forced '%s'", &proxy_request[url_start]);
+#endif
+#ifdef DEBUG
+			fprintf(stderr, "Forced '%s'\n", &proxy_request[url_start]);
+#endif
+		}
+		else if (translate_url)
+		{
+#ifdef LOG_TO_SYSLOG
+			syslog(LOG_NOTICE, "Translated '%s'", &proxy_request[url_start]);
+#endif
+#ifdef DEBUG
+			fprintf(stderr, "Translated '%s'\n", &proxy_request[url_start]);
+#endif
+		}
 		else
 		{
-			write(proxy, headers, header_len);
-
-			/*
-			 * Dump a nice log message showing what we have done.
-			 */
-			if ((uri_pos = strchr(headers, ' ')) != NULL)
-			{
-				if ((uri_end = strchr(uri_pos + 1, ' ')) != NULL)
-				{
-					*uri_end = '\0';
-					syslog(LOG_DEBUG, "Untranslated '%s'", headers);
+#ifdef LOG_TO_SYSLOG
+			syslog(LOG_NOTICE, "Untranslated '%s'", &proxy_request[url_start]);
+#endif
 #ifdef DEBUG
-					fprintf(stderr, "Untranslated '%s'\n", headers);
+			fprintf(stderr, "Untranslated '%s'\n", &proxy_request[url_start]);
 #endif
-				}
-			}
 		}
 	}
 
@@ -838,7 +1034,9 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 		 */
 		if (select(max_fd + 1, &read_fd, NULL, NULL, NULL) < 0)
 		{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 			syslog(LOG_ERR, "select(): %m");
+#endif
 			close(proxy);
 			return;
 		}
@@ -856,21 +1054,28 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 			switch (read_len = read(sock, headers, sizeof(headers) - 1))
 			{
 			case -1:
-				syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_WARNING),
-						"read(client) failed: %m");
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+				if (errno != ECONNRESET)
+					syslog(LOG_WARNING, "read(client) failed: %m");
+#ifdef LOG_TO_SYSLOG
+				else
+					syslog(LOG_DEBUG, "read(client) failed: %m");
+#endif
+#endif
 				close(proxy);
 				return;
 
 			case 0:
-				close(sock);
 				close(proxy);
 				return;
 
 			default:
 				if (write(proxy, headers, read_len) < 0)
 				{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 					syslog(LOG_WARNING, "write(proxy) failed: %m");
-					close(sock);
+#endif
+					close(proxy);
 					return;
 				}
 				break;
@@ -885,20 +1090,25 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 			switch (read_len = read(proxy, headers, sizeof(headers) - 1))
 			{
 			case -1:
-				syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_WARNING),
-						"read(proxy) failed: %m");
-				close(sock);
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+				if (errno != ECONNRESET)
+					syslog(LOG_WARNING, "read(proxy) failed: %m");
+				else
+					syslog(LOG_DEBUG, "read(proxy) failed: %m");
+#endif
+				close(proxy);
 				return;
 
 			case 0:
 				close(proxy);
-				close(sock);
 				return;
 
 			default:
 				if (write(sock, headers, read_len) < 0)
 				{
+#ifdef LOG_TO_SYSLOG
 					syslog(LOG_WARNING, "write(client) failed: %m");
+#endif
 					close(proxy);
 					return;
 				}
@@ -933,8 +1143,14 @@ static int	get_site_from_headers(int sock, char *hostname, int hostlen,
 	switch (read_len = read(sock, headers, header_max))
 	{
 	case -1:
-		syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_WARNING),
-				"read(client) failed: %m");
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+		if (errno != ECONNRESET)
+			syslog(LOG_WARNING, "read(client) failed: %m");
+#ifdef LOG_TO_SYSLOG
+		else
+			syslog(LOG_DEBUG, "read(client) failed: %m");
+#endif
+#endif
 		return (-1);
 
 	case 0:
@@ -991,6 +1207,60 @@ static int	get_site_from_headers(int sock, char *hostname, int hostlen,
  */
 static void alarm_signal(int sig)
 {
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 	syslog(LOG_DEBUG, "Alarm signal caught - connection timeout");
+#endif
 	exit(1);
 }
+
+/*
+ * Appends source string to destination string.
+ */
+static void append_string(char *dest, int max_size, char *source)
+{
+	size_t	dest_len;
+	size_t	source_len;
+	size_t	max_add_len;
+ 
+ 	/*
+	 * Get the length of what is already there, and what we have to add.
+	 */
+	dest_len = strlen(dest);
+	source_len = strlen(source);
+ 
+ 	/*
+	 * Calculate how many characters we can append.
+	 */
+ 	if ((dest_len + source_len + 1) < max_size)
+		max_add_len = source_len;
+	else
+		max_add_len = max_size - dest_len - 1;
+
+	/*
+	 * If any space at all available, then append what we can.
+	 */
+	if (max_add_len > 0)
+	{
+		memcpy(&dest[dest_len], source, max_add_len);
+		dest[dest_len + max_add_len] = '\0';
+	}
+}
+
+static void append_char(char *dest, int max_size, char source)
+{
+	size_t	dest_len;
+ 
+ 	/*
+	 * Get the length of what is already there.
+	 */
+	dest_len = strlen(dest);
+ 
+	/*
+	 * If any space at all available, then append what we can.
+	 */
+	if ((dest_len + 1) < max_size)
+	{
+		dest[dest_len] = source;
+		dest[dest_len + 1] = '\0';
+	}
+}
